class Keywords:
    COMMAND = "dls_servbase_guis::keywords::command"
    PAYLOAD = "dls_servbase_guis::keywords::payload"
    TAB_ID = "dls_servbase_guis::keywords::tab_id"


class Commands:
    LOAD_TABS = "dls_servbase_guis::commands::load_tabs"
    SELECT_TAB = "dls_servbase_guis::commands::select_tab"


class Cookies:
    TABS_MANAGER = "dls_servbase_guis::cookies::tabs_manager"
