class Keywords:
    COMMAND = "dls_servbase_datafaces::keywords::command"
    PAYLOAD = "dls_servbase_datafaces::keywords::payload"


class Commands:
    EXECUTE = "dls_servbase_datafaces::commands::execute"
